"""
Abstract synthesis super-class.

Users should not interact with this file, but any concrete synthesis methods should
inherit one of these classes, to provide a unified interface.
"""

import abc
import importlib
import inspect
import math
import warnings
from typing import Any, Literal

import numpy as np
import torch

from ..tools import examine_saved_synthesis
from ..tools.data import _check_tensor_equality


def _get_name(x: object) -> str:
    """
    Get the name of an object ``x``, for saving/loading purposes.

    Parameters
    ----------
    x
        A python object or function.

    Returns
    -------
    name
        The name of that object, with full module path (e.g.,
        ``torch.optim.optimizer.Adam``).
    """  # numpydoc ignore=ES01
    if x is None:
        return None
    try:
        # if this passes, attr is a function
        name = f"{x.__module__}.{x.__name__}"
    except AttributeError:
        # if we're here, then it's an object
        cls = x.__class__
        name = f"{cls.__module__}.{cls.__name__}"
    return name


class Synthesis(abc.ABC):
    r"""
    Abstract super-class for synthesis objects.

    All synthesis objects share a variety of similarities and thus need
    to have similar methods. Some of these can be implemented here and
    simply inherited, some of them will need to be different for each
    sub-class and thus are marked as abstract methods here.
    """

    def __init__(self):
        # flag to raise more informative error message if setup and load are both called
        self._loaded = False

    @abc.abstractmethod
    def synthesize(self):
        r"""Synthesize something."""  # numpydoc ignore=ES01
        pass

    def save(
        self,
        file_path: str,
        save_io_attrs: list[tuple[str]] = [],
        save_state_dict_attrs: list[str] = [],
    ):
        r"""
        Save all attributes in .pt file.

        Note that there are two special categories of attributes, as described below.
        All other attributes will be pickled directly and so should be either tensors or
        primitives (e.g., not a function or callable torch object). We do not check this
        explicitly, but load will fail if that's not the case.

        Parameters
        ----------
        file_path
            The path to save the synthesis object to.
        save_io_attrs
            List with tuples of form (str, (str, ...)). The first element is the name of
            the attribute to we save, and the second element is a tuple of attributes of
            the Synthesis object, which we can pass as inputs to the attribute. We save
            them as tuples of (name, input_names, outputs). On load, we check that the
            initialized object's name hasn't changed, and that when called on the same
            inputs, we get the same outputs. Intended for models, metrics, loss
            functions. Used to avoid saving callables, which is brittle and unsafe.
        save_state_dict_attrs
            Names of attributes that we save as tuples of (name, state_dict).
            Corresponding attribute can be None, in which case we save an empty
            dictionary as state_dict. On load, we check that the initialized object's
            name hasn't changed, and load the state_dict. Intended for optimizers,
            schedulers. Used to avoid saving callables, which is brittle and unsafe.

        Raises
        ------
        ValueError
            If any of the strings specified in ``save_io_attrs`` as inputs for the
            callable attribute-to-save are not attributes of ``self``.
        """
        save_dict = {}
        save_dict["save_metadata"] = {
            "plenoptic_version": importlib.metadata.version("plenoptic"),
            "torch_version": importlib.metadata.version("torch"),
            "synthesis_object": _get_name(self),
        }
        save_io_attr_names = [k[0] for k in save_io_attrs]
        save_attrs = [
            k for k in vars(self) if k not in save_io_attr_names + save_state_dict_attrs
        ]
        for k in save_attrs:
            attr = getattr(self, k)
            # detaching the tensors avoids some headaches like the
            # tensors having extra hooks or the like
            if isinstance(attr, torch.Tensor):
                attr = attr.detach()
            save_dict[k] = attr
        for k, input_names in save_io_attrs:
            attr = getattr(self, k)
            name = _get_name(attr)
            tensors = [getattr(self, t) for t in input_names]
            save_dict[k] = (name, input_names, attr(*tensors))
            if any([n not in save_dict for n in input_names]):
                raise ValueError(
                    "input_name must be included in save dictionary, "
                    f"but got {input_names}!"
                )
        for k in save_state_dict_attrs:
            attr = getattr(self, k)
            name = _get_name(attr)
            try:
                state_dict = attr.state_dict()
            except AttributeError:
                # then we assume that attr is None
                state_dict = {}
            save_dict[k] = (name, state_dict)
        torch.save(save_dict, file_path)

    def load(
        self,
        file_path: str,
        empty_on_init_attr: str,
        map_location: str | None = None,
        check_attributes: list[str] = [],
        check_io_attributes: list[str] = [],
        state_dict_attributes: list[str] = [],
        tensor_equality_atol: float = 1e-8,
        tensor_equality_rtol: float = 1e-5,
        **pickle_load_args: Any,
    ):
        r"""
        Load all relevant attributes from a .pt file.

        This should be called by ``Synthesis`` object that has just been initialized.

        Note this operates in place and so doesn't return anything.

        Parameters
        ----------
        file_path
            The path to load the synthesis object from.
        empty_on_init_attr
            The name of an attribute that will either be None or have length 0 if the
            Synthesis object has just been initialized.
        map_location
            Argument to pass to :func:`torch.load` as ``map_location``. If you save
            stuff that was being run on a GPU and are loading onto a
            CPU, you'll need this to make sure everything lines up
            properly. This should be structured like the str you would
            pass to :class:`torch.device`.
        check_attributes
            List of strings we ensure are identical in the current ``Synthesis`` object
            and the loaded one.
        check_io_attributes
            Names of attributes whose input/output behavior we should check (i.e., if we
            call them on identical inputs, do we get identical outputs). In the loaded
            dictionary, these are a tuple of three values: the name of the callable, the
            name of the attribute to use as input, and the output we expect.
        state_dict_attributes
            Names of attributes that were callables, saved as a tuple with the name of
            the callable and their state_dict. We will ensure the name of the attributes
            are identical and then load the state_dict. If the attribute is None on the
            initialized Synthesis object, then we set the tuple, and count on the
            Synthesis object to properly handle it when needed.
        tensor_equality_atol
            Absolute tolerance to use when checking for tensor equality during load,
            passed to :func:`torch.allclose`. It may be necessary to increase if you are
            saving and loading on two machines with torch built by different cuda
            versions. Be careful when changing this! See
            :class:`torch.finfo<torch.torch.finfo>` for more details about floating
            point precision of different data types (especially, ``eps``); if you have
            to increase this by more than 1 or 2 decades, then you are probably not
            dealing with a numerical issue.
        tensor_equality_rtol
            Relative tolerance to use when checking for tensor equality during load,
            passed to :func:`torch.allclose`. It may be necessary to increase if you are
            saving and loading on two machines with torch built by different cuda
            versions. Be careful when changing this! See
            :class:`torch.finfo<torch.torch.finfo>` for more details about floating
            point precision of different data types (especially, ``eps``); if you have
            to increase this by more than 1 or 2 decades, then you are probably not
            dealing with a numerical issue.
        **pickle_load_args
            Any additional kwargs will be added to ``pickle_module.load`` via
            :func:`torch.load`, see that function's docstring for details.

        Raises
        ------
        ValueError
            If the loading object has not just been initialized.
        ValueError
            If the object saved at ``file_path`` is not the same type as the loading
            object.
        ValueError
            If either the saved or loading object has attributes not found in the
            other.
        ValueError
            If the saved and loading objects have a different value for one of the
            ``check_attributes``.
        ValueError
            If the behavior of one of the ``check_io_attributes`` is different between
            the saved and loading objects.

        Warns
        -----
        UserWarning
            If :func:`setup` will need to be called after load, to finish initializing
            one of the ``state_dict_attributes``
        """
        check_str = (
            "\n\nIf this is confusing, try calling "
            f"{_get_name(examine_saved_synthesis)}('{file_path}'),"
            " to examine saved object"
        )
        empty_on_init_attr = getattr(self, empty_on_init_attr)
        if empty_on_init_attr is not None and len(empty_on_init_attr) > 0:
            raise ValueError(
                "load can only be called with a just-initialized"
                f" {self.__class__.__name__} object"
            )
        tmp_dict = torch.load(
            file_path,
            map_location=map_location,
            **pickle_load_args,
        )
        metadata = tmp_dict.pop("save_metadata")
        if metadata["synthesis_object"] != _get_name(self):
            raise ValueError(
                f"Saved object was a {metadata['synthesis_object']}"
                f", but initialized object is {_get_name(self)}! "
                f"{check_str}"
            )
        # all attributes set at initialization should be present in the saved dictionary
        init_not_save = set(vars(self)) - set(tmp_dict)
        if len(init_not_save):
            # in PR #370 (release 1.3.1), added _current_loss attribute, which we'll
            # handle for now, but warn about.
            if init_not_save == {"_current_loss"}:
                tmp_dict["_current_loss"] = None
                warnings.warn(
                    "The saved object was saved with plenoptic 1.3.0 or earlier and "
                    "will not be compatible with future releases. Save this object "
                    "with current version of plenoptic or see the 'Reproducibility "
                    "and Compatibility' page of the documentation for how to make the "
                    "saved object futureproof and avoid this warning.",
                    category=FutureWarning,
                )
            else:
                init_not_save_str = "\n ".join(
                    [f"{k}: {getattr(self, k)}" for k in init_not_save]
                )
                raise ValueError(
                    f"Initialized object has {len(init_not_save)} attribute(s) "
                    f"not present in the saved object!\n {init_not_save_str}"
                )
        # there shouldn't be any extra keys in the saved dictionary (we removed
        # save_metadata above)
        save_not_init = set(tmp_dict) - set(vars(self))
        if len(save_not_init):
            save_not_init_str = "\n ".join(
                [f"{k}: {tmp_dict[k]}" for k in save_not_init]
            )
            raise ValueError(
                f"Saved object has {len(save_not_init)} attribute(s) "
                f"not present in the initialized object!\n {save_not_init_str}"
            )
        for k in check_attributes:
            # The only hidden attributes we'd check are those like
            # range_penalty_lambda, where this function is checking the
            # hidden version (which starts with '_'), but during
            # initialization, the user specifies the version without
            # the initial underscore. This is because this function
            # needs to be able to set the attribute, which can only be
            # done with the hidden version.
            display_k = k[1:] if k.startswith("_") else k
            if isinstance(getattr(self, k), torch.Tensor):
                _check_tensor_equality(
                    tmp_dict[k],
                    getattr(self, k),
                    "Saved",
                    "Initialized",
                    error_prepend_str=(
                        f"Saved and initialized attribute {display_k} have "
                        f"different {{error_type}}!"
                    ),
                    error_append_str=check_str,
                    atol=tensor_equality_atol,
                    rtol=tensor_equality_rtol,
                )
            elif isinstance(getattr(self, k), float):
                if not np.allclose(getattr(self, k), tmp_dict[k]):
                    raise ValueError(
                        f"Saved and initialized {display_k} are different!"
                        f"\nSaved: {tmp_dict[k]}"
                        f"\nInitialized: {getattr(self, k)}"
                        f"{check_str}"
                    )
            else:
                if getattr(self, k) != tmp_dict[k]:
                    raise ValueError(
                        f"Saved and initialized {display_k} are different!"
                        f"\nSaved: {tmp_dict[k]}"
                        f"\nInitialized: {getattr(self, k)}"
                        f"{check_str}"
                    )
        for k, input_names in check_io_attributes:
            # same as above
            display_k = k[1:] if k.startswith("_") else k
            tensors = [tmp_dict[t] for t in tmp_dict[k][1]]
            init_name = _get_name(getattr(self, k))
            saved_name = tmp_dict[k][0]
            init_loss = getattr(self, k)(*tensors)
            saved_loss = tmp_dict[k][-1]
            _check_tensor_equality(
                saved_loss,
                init_loss,
                f"Saved ({saved_name})",
                f"Initialized ({init_name})",
                error_prepend_str=(
                    f"Saved and initialized {display_k} output have "
                    f"different {{error_type}}!"
                ),
                error_append_str=check_str,
                atol=tensor_equality_atol,
                rtol=tensor_equality_rtol,
            )
        for k, v in tmp_dict.items():
            # check_io_attributes is a tuple
            if k in [a[0] for a in check_io_attributes] + state_dict_attributes:
                display_k = k[1:] if k.startswith("_") else k
                init_attr = getattr(self, k, None)
                # then check their name has the same final part (e.g.,
                # "plenoptic.simulate.PortillaSimoncelli" or
                # "__main__.PortillaSimoncelli"), since we've already checked
                # the behavior, keep going (don't update the object's attribute)
                if init_attr is not None:
                    init_name = _get_name(init_attr)
                    saved_name = v[0]
                    init_name = init_name.split(".")[-1]
                    saved_name = saved_name.split(".")[-1]
                    if init_name != saved_name:
                        raise ValueError(
                            f"Saved and initialized {display_k} "
                            "have different names!"
                            f"\nSaved: {saved_name}"
                            f"\nInitialized: {init_name}"
                            f"{check_str}"
                        )
                    continue
                # if init_attr is None, then we haven't set it yet, so we set the saved
                # tuple as the attribute, and handle this later
            setattr(self, k, v)
        setup_attrs = []
        optim = tmp_dict.get("_optimizer", None)
        if isinstance(optim, tuple) and optim[0] != _get_name(torch.optim.Adam):
            setup_attrs.append("optimizer")
        sched = tmp_dict.get("_scheduler", None)
        if isinstance(sched, tuple) and sched[0] is not None:
            setup_attrs.append("scheduler")
        if setup_attrs:
            warnings.warn(
                f"You will need to call setup() to instantiate {', '.join(setup_attrs)}"
            )
        # Make sure we specify that we have loaded and setup has not been called
        self._loaded = True

    @abc.abstractmethod
    def to(self, *args: Any, attrs: list[str] = [], **kwargs: Any):
        r"""
        Move and/or cast the parameters and buffers.

        Similar to :func:`save`, this is an abstract method only because you
        need to define the attributes to call to on.

        This can be called as

        .. function:: to(device=None, dtype=None, non_blocking=False)

        .. function:: to(dtype, non_blocking=False)

        .. function:: to(tensor, non_blocking=False)

        Its signature is similar to :meth:`torch.Tensor.to`, but only accepts
        floating point desired :attr:`dtype` s. In addition, this method will
        only cast the floating point parameters and buffers to :attr:`dtype`
        (if given). The integral parameters and buffers will be moved
        :attr:`device`, if that is given, but with dtypes unchanged. When
        :attr:`non_blocking` is set, it tries to convert/move asynchronously
        with respect to the host if possible, e.g., moving CPU Tensors with
        pinned memory to CUDA devices.

        When calling this method to move tensors to a CUDA device, items in ``attrs``
        that start with ``"saved_"`` will not be moved.

        See :meth:`torch.nn.Module.to` for examples.

        .. note::
            This method modifies the module in-place.

        Parameters
        ----------
        device : torch.device
            The desired device of the parameters and buffers in this module.
        dtype : torch.dtype
            The desired floating point type of the floating point parameters and
            buffers in this module.
        tensor : torch.Tensor
            Tensor whose dtype and device are the desired dtype and device for
            all parameters and buffers in this module.
        """  # numpydoc ignore=PR01,PR02
        device, dtype, non_blocking, memory_format = torch._C._nn._parse_to(
            *args, **kwargs
        )

        def move(a: torch.Tensor, k: str) -> torch.Tensor:
            # numpydoc ignore=RT01,ES01,PR01,GL08
            move_device = None if k.startswith("saved_") else device
            if memory_format is not None and a.dim() == 4:
                return a.to(
                    move_device,
                    dtype,
                    non_blocking,
                    memory_format=memory_format,
                )
            else:
                return a.to(move_device, dtype, non_blocking)

        for k in attrs:
            if hasattr(self, k):
                attr = getattr(self, k)
                if isinstance(attr, torch.Tensor):
                    attr = move(attr.data, k)
                    if isinstance(getattr(self, k), torch.nn.Parameter):
                        attr = torch.nn.Parameter(attr)
                    if getattr(self, k).requires_grad:
                        attr = attr.requires_grad_()
                    setattr(self, k, attr)
                elif isinstance(attr, list):
                    setattr(self, k, [move(a, k) for a in attr])
                elif attr is not None:
                    setattr(self, k, move(attr, k))


class OptimizedSynthesis(Synthesis):
    r"""
    Abstract super-class for synthesis objects that use optimization.

    The primary difference between this and the generic Synthesis class is that
    these will use an optimizer object to iteratively update their output.

    Parameters
    ----------
    range_penalty_lambda
        Strength of the regularizer that enforces the allowed_range. Must be
        non-negative.
    allowed_range
        Range (inclusive) of allowed pixel values. Any values outside this
        range will be penalized.
    """

    def __init__(
        self,
        range_penalty_lambda: float = 0.1,
        allowed_range: tuple[float, float] = (0, 1),
    ):
        super().__init__()
        self._losses = []
        self._gradient_norm = []
        self._pixel_change_norm = []
        self._store_progress = None
        self._optimizer = None
        self._current_loss = None
        if range_penalty_lambda < 0:
            raise Exception("range_penalty_lambda must be non-negative!")
        self._range_penalty_lambda = range_penalty_lambda
        self._allowed_range = allowed_range

    @abc.abstractmethod
    def setup(self):
        r"""Initialize relevant attributes."""  # numpydoc ignore=ES01
        pass

    @abc.abstractmethod
    def objective_function(self):
        r"""
        How good is the current synthesized object.

        See ``plenoptic.tools.optim`` for some examples.
        """
        pass

    @abc.abstractmethod
    def _check_convergence(self):
        r"""
        How to determine if synthesis has finished.

        See ``plenoptic.tools.convergence`` for some examples.
        """
        pass

    def _closure(self) -> torch.Tensor:
        r"""
        Calculate the gradient, before the optimization step.

        This enables optimization algorithms that perform several evaluations
        of the gradient before taking a step (ie. second order methods like
        LBFGS).

        Additionally, this is where ``loss`` is calculated and
        ``loss.backward()`` is called.

        Returns
        -------
        loss
            Loss of the current objective function.
        """
        self.optimizer.zero_grad()
        loss = self.objective_function()
        loss.backward(retain_graph=False)
        return loss.item()

    def _initialize_optimizer(
        self,
        optimizer: torch.optim.Optimizer | None,
        synth_attr: torch.Tensor,
        optimizer_kwargs: dict | None = None,
        learning_rate: float = 0.01,
    ):
        """
        Initialize optimizer.

        optimizer can be:

        - None, in which case we create an Adam optimizer with amsgrad=True and
          ``lr=learning_rate``

        - uninitialized torch.optim.Optimizer, in which case it must already have the
          synthesis attribute (e.g., metamer) as its only parameter.

        In either case, we then initialize the optimizer with a single parameter, the
        synthesis attribute, ``synth_attr``

        If we have loaded from a save state, self.optimizer will be a tuple with the
        name of the optimizer class (e.g., torch.optim.adam.Adam) and the
        optimizer's state_dict. In that case:

        - If optimizer is None, the saved optimizer must be Adam, and we load the
          state_dict.

        - else, the saved and user-specified optimizers must have the same class name
          and we load the state_dict. in this case, ``optimizer_kwargs`` must be None

        Parameters
        ----------
        optimizer
            The (un-initialized) optimizer object to use. If ``None``, we use
            :class:`torch.optim.Adam`.
        synth_attr
            The Tensor that we are updating as part of optimization.
        optimizer_kwargs
            The keyword arguments to pass to the optimizer on initialization. If
            ``None``, we use ``{"lr": .01}`` and, if optimizer is ``None``,
            ``{"amsgrad": True}``.
        learning_rate
            The learning rate for the optimizer.

        Raises
        ------
        ValueError
            If ``optimizer_kwargs`` is not ``None`` and ``self.optimizer`` is a tuple
            (thus, :meth:`load` was called before this).
        TypeError
            If ``self.optimizer`` is a tuple (thus, :meth:`load` was called before
            this), was not :class`torch.optim.Adam`, and ``optimizer`` arg is ``None``.
        ValueError
            If ``self.optimizer`` is a tuple (thus, :meth:`load` was called before this)
            but ``optimizer`` arg has a different type than the saved optimizer.
        """
        if isinstance(self.optimizer, tuple):
            # then we're calling this after load()
            if optimizer_kwargs is not None:
                raise ValueError(
                    "When initializing optimizer after load, optimizer_kwargs"
                    " must be None!"
                )
            if optimizer is None:
                optimizer = torch.optim.Adam
                err_str = (
                    "Don't know how to initialize saved optimizer "
                    f"'{self._optimizer[0]}'! Pass an un-initialized version of "
                    "this optimizer to `setup`, and we will update its "
                    "state_dict."
                )
            else:
                err_str = (
                    "User-specified optimizer must have same type as saved "
                    "optimizer, but got:"
                    f"\nSaved: {self.optimizer[0]}, "
                    f"\nUser-specified: {_get_name(optimizer)}."
                )

            if self._optimizer[0] != _get_name(optimizer):
                raise ValueError(err_str)
            state_dict = self.optimizer[1]
            self._optimizer = optimizer([synth_attr])
            self._optimizer.load_state_dict(state_dict)
        else:
            if optimizer_kwargs is None:
                optimizer_kwargs = {"lr": learning_rate}
            else:
                optimizer_kwargs.setdefault("lr", learning_rate)
            if optimizer is None:
                optimizer_kwargs.setdefault("amsgrad", True)
                self._optimizer = torch.optim.Adam([synth_attr], **optimizer_kwargs)
            else:
                self._optimizer = optimizer([synth_attr], **optimizer_kwargs)

    def _initialize_scheduler(
        self,
        scheduler: torch.optim.lr_scheduler.LRScheduler | None,
        optimizer: torch.optim.Optimizer,
        scheduler_kwargs: dict | None = None,
    ):
        """
        Initialize scheduler.

        scheduler can be:

        - None, in which case we do nothing.

        - uninitialized torch.optim.lr_scheduler.LRScheduler in which case we initialize
          it using self.optimizer.

        If we have loaded from a save state, self.scheduler will be a tuple with the
        name of the scheduler class (e.g., torch.optim.lr_scheduler.ConstantLR) and the
        scheduler's state_dict. In that case:

        - If scheduler is None, we raise a ValueError.

        - else, the saved and user-specified schedulers must have the same class name
          and we load the state_dict. In this case, ``scheduler_kwargs`` must be None

        We also check if scheduler.step takes any arguments beyond self and epoch. If
        so, we set the `self._scheduler_step_arg` flag to True (it was set to False at
        object initialization). Then, we will pass the loss to scheduler.step when we
        call it.

        Parameters
        ----------
        scheduler
            The un-initialized learning rate scheduler object to use. If ``None``, we
            don't use one.
        optimizer
            The initialized optimizer whose learning rate ``scheduler`` will be
            adjusting.
        scheduler_kwargs
            The keyword arguments to pass to the scheduler on initialization.

        Raises
        ------
        ValueError
            If ``scheduler_kwargs`` is not ``None`` and ``self.scheduler`` is a tuple
            (thus, :meth:`load` was called before this).
        TypeError
            If ``self.scheduler`` is a tuple (thus, :meth:`load` was called before this)
            but ``scheduler`` arg is ``None``.
        ValueError
            If ``self.scheduler`` is a tuple (thus, :meth:`load` was called before this)
            but ``scheduler`` arg has a different type than the saved scheduler.
        """
        if isinstance(self.scheduler, tuple):
            # then we're calling this after load()
            if scheduler_kwargs is not None:
                raise ValueError(
                    "When initializing scheduler after load, scheduler_kwargs"
                    " must be None!"
                )
            if scheduler is None:
                if self.scheduler[0] is not None:
                    # then this comes from loading, and we don't know how to initialize
                    # the scheduler
                    raise TypeError(
                        "Don't know how to initialize saved scheduler "
                        f"'{self.scheduler[0]}'! Pass an initialized version of this "
                        "scheduler to `synthesize`, and we will update its state_dict."
                    )
                self._scheduler = None
            else:
                if self.scheduler[0] != _get_name(scheduler):
                    raise ValueError(
                        "User-specified scheduler must have same type as saved "
                        f"scheduler, but got:\nSaved: {self.scheduler[0]}, "
                        f"\nUser-specified: {_get_name(scheduler)}."
                    )
                state_dict = self.scheduler[1]
                self._scheduler = scheduler(optimizer)
                self._scheduler.load_state_dict(state_dict)
        else:
            if scheduler_kwargs is None:
                scheduler_kwargs = {}
            if scheduler is not None:
                self._scheduler = scheduler(optimizer, **scheduler_kwargs)
                step_args = set(inspect.getfullargspec(self.scheduler.step).args)
                if len(step_args - {"self", "epoch"}):
                    # then we do want to pass the loss to scheduler.step
                    self._scheduler_step_arg = True

    def _convert_iteration(
        self,
        iteration: int,
        iteration_to_progress: bool = True,
        iteration_selection: Literal["floor", "ceiling", "round"] = "round",
    ) -> int:
        """
        Convert between synthesis iteration and ``store_progress``'s iteration.

        Several of the ``OptimizedSynthesis`` attributes are not updated every
        single iteration but every ``self.store_progress`` iterations. This
        converts between the two.

        Parameters
        ----------
        iteration
            Synthesis iteration to summarize.
        iteration_to_progress
            Whether to convert from synthesis iteration to ``store_progress``
            iteration (in which case, behavior is controlled by
            ``iteration_selection``) or vice-versa (in which case we return
            ``iteration * self.store_progress``).
        iteration_selection

            How to select the relevant iteration from the saved synthesis attribute
            when the request iteration wasn't stored.

            When synthesis was run with ``store_progress=n`` (where ``n>1``),
            synthesis outputs are only saved every ``n`` iterations. If you request an
            iteration where synthesis wasn't saved, this determines which available
            iteration is used instead:

            * ``"floor"``: use the closest saved iteration **before** the
              requested one.

            * ``"ceiling"``: use the closest saved iteration **after** the
              requested one.

            * ``"round"``: use the closest saved iteration.

        Returns
        -------
        converted_iteration
            Converted iteration.
        """
        if iteration_to_progress:
            # round and ceiling may be one greater than e.g., len(self._saved_metamer).
            # however, self.saved_metamer always has the current metamer appended, and
            # so this will be okay
            if iteration_selection == "floor":
                iter = math.floor(iteration / self.store_progress)
            elif iteration_selection == "round":
                iter = round(iteration / self.store_progress)
            elif iteration_selection == "ceiling":
                iter = math.ceil(iteration / self.store_progress)
        else:
            iter = iteration * self.store_progress
            # this might go off the end
            if iter >= len(self.losses):
                iter = len(self.losses) - 1
        return iter

    @abc.abstractmethod
    def get_progress(
        self,
        iteration: int | None,
        iteration_selection: Literal["floor", "ceiling", "round"] = "round",
        addt_every_iter_attributes: list[str] = [],
        store_progress_attributes: list[str] = [],
    ) -> dict[str, torch.Tensor | None | int]:
        """
        Return dictionary summarizing synthesis progress at ``iteration``.

        Note that for the most recent iteration (``iteration=-1`` or ``iteration=None``
        or ``iteration==len(self.losses)-1``), we do not have values for
        :attr:`pixel_change_norm`, :attr:`gradient_norm` or
        ``addt_every_iter_attributes``, since in this case we are showing the loss and
        value for the current synthesis output.

        Parameters
        ----------
        iteration
            Synthesis iteration to summarize. If ``None``, grab the most recent.
            Negative values are allowed.
        iteration_selection

            How to select the relevant iteration from the saved synthesis attribute
            when the request iteration wasn't stored.

            When synthesis was run with ``store_progress=n`` (where ``n>1``),
            synthesis outputs are only saved every ``n`` iterations. If you request an
            iteration where synthesis wasn't saved, this determines which available
            iteration is used instead:

            * ``"floor"``: use the closest saved iteration **before** the
              requested one.

            * ``"ceiling"``: use the closest saved iteration **after** the
              requested one.

            * ``"round"``: use the closest saved iteration.

        addt_every_iter_attributes
            Additional attributes that have values appended every iteration.
            ``losses``, ``pixel_change_norm`` and ``gradient_norm`` are always
            included.
        store_progress_attributes
            Attributes that have values appended every ``self.store_progress``
            iterations (and may thus be ``None``).

        Returns
        -------
        progress_info
            Dictionary summarizing synthesis progress.

        Raises
        ------
        IndexError
            If ``iteration`` takes an illegal value.

        Warns
        -----
        UserWarning
            If the iteration used for ``store_progress_attributes`` is not the same as
            the argument ``iteration`` (because e.g., you set ``iteration=3`` but
            ``self.store_progress=2``).
        """
        if iteration is None:
            iter = len(self.losses) - 1
        elif iteration < 0:
            iter = len(self.losses) + iteration
        else:
            iter = iteration

        if iter < 0:
            # if this is negative after our remapping above, then it was too large
            # and it wrapped around again (e.g., -100 when there were only 90
            # iterations)
            raise IndexError(
                f"{iteration=} out of bounds with "
                f"{len(self.losses)} iterations of synthesis"
            )

        # len(self.losses) is the number of synthesis iterations plus 1 (for the current
        # loss), so we grab the hidden version, which has the proper length
        try:
            loss = self.losses[iter]
        except IndexError as e:
            raise IndexError(
                f"{iteration=} out of bounds with "
                f"{len(self.losses)} iterations of synthesis"
            ) from e
        progress_info = {"losses": loss, "iteration": iter}
        # then this is the most recent one, which we don't have pixel_change_norm or
        # gradient_norm for
        if iter == len(self.losses) - 1:
            progress_info.update({"pixel_change_norm": None, "gradient_norm": None})
        else:
            progress_info.update(
                {
                    "pixel_change_norm": self.pixel_change_norm[iter],
                    "gradient_norm": self.gradient_norm[iter],
                }
            )
        progress_info.update(
            {k: getattr(self, k)[iter] for k in addt_every_iter_attributes}
        )
        if self.store_progress:
            # treat None special: always grab current one
            if iteration is None:
                store_progress_iter = -1
            else:
                store_progress_iter = self._convert_iteration(
                    iter, iteration_selection=iteration_selection
                )
            progress_info.update(
                {
                    k: getattr(self, k)[store_progress_iter]
                    for k in store_progress_attributes
                }
            )
            # treat None special: always grab current one
            if iteration is None:
                store_progress_iter = iter
            else:
                store_progress_iter = self._convert_iteration(
                    store_progress_iter, False, iteration_selection
                )
            if store_progress_iter != iter:
                warnings.warn(
                    f"loss iteration and iteration for {store_progress_attributes} are"
                    " not the same"
                )
            progress_info.update({"store_progress_iteration": store_progress_iter})
        return progress_info

    @property
    def range_penalty_lambda(self) -> float:
        """Magnitude of the penalty on pixel values outside :attr:`allowed_range`."""
        # numpydoc ignore=RT01,ES01
        return self._range_penalty_lambda

    @property
    def allowed_range(self) -> tuple[float, float]:
        """Allowable range of pixel values."""
        # numpydoc ignore=RT01,ES01
        return self._allowed_range

    @property
    def losses(self) -> torch.Tensor:
        """
        Optimization loss over iterations.

        Will have ``length=num_iter+1``, where ``num_iter`` is the number of
        iterations of synthesis run so far.

        This tensor always lives on the CPU.
        """  # numpydoc ignore=RT01
        current_loss = self._current_loss
        # this will happen if we haven't run synthesize() yet or got
        # interrupted
        if current_loss is None:
            try:
                # compute current loss, no need to compute gradient
                with torch.no_grad():
                    current_loss = self.objective_function().item()
            except RuntimeError as e:
                exp_msg = "a Tensor with 0 elements cannot be converted to Scalar"
                if e.args[0] != exp_msg:
                    raise e
                # this will happen if setup() has not been called and so we can't
                # compute loss because synthesis hasn't been initialized.
                return torch.empty(0)
        return torch.as_tensor([*self._losses, current_loss])

    @property
    def gradient_norm(self) -> torch.Tensor:
        """Optimization gradient's L2 norm over iterations."""
        # numpydoc ignore=RT01,ES01
        return torch.as_tensor(self._gradient_norm)

    @property
    def pixel_change_norm(self) -> torch.Tensor:
        """L2 norm change in pixel values over iterations."""
        # numpydoc ignore=RT01,ES01
        return torch.as_tensor(self._pixel_change_norm)

    @property
    def store_progress(self) -> bool | int:
        """
        How often we are caching progress.

        If ``False``, we don't save anything. If ``True``, we save every iteration. If
        an int, we save every ``store_progress`` iterations (note then that 0 is the
        same as ``False`` and 1 the same as ``True``).
        """  # numpydoc ignore=RT01
        return self._store_progress

    @store_progress.setter
    def store_progress(self, store_progress: bool | int):
        """
        Initialize store_progress.

        Parameters
        ----------
        store_progress : bool or int, optional
            Whether we should store the synthesis output in progress on every
            iteration. If ``False``, we don't save anything. If ``True``, we save
            every iteration. If an int, we save every ``store_progress`` iterations
            (note then that 0 is the same as ``False`` and 1 the same as ``True``).

        Raises
        ------
        ValueError
            If ``store_progress`` has already been set and you are trying to change the
            value.
        """
        # numpydoc ignore=RT01,ES01
        if store_progress and store_progress is True:
            store_progress = 1
        if self.store_progress is not None and store_progress != self.store_progress:
            # we require store_progress to be the same because otherwise the
            # subsampling relationship between attrs that are stored every
            # iteration (loss, gradient, etc) and those that are stored every
            # store_progress iteration (e.g., saved_metamer) changes partway
            # through and that's annoying
            raise ValueError(
                "If you've already run synthesize() before, must "
                "re-run it with same store_progress arg. You "
                f"passed {store_progress} instead of "
                f"{self.store_progress} (True is equivalent to 1)"
            )
        self._store_progress = store_progress

    @property
    def optimizer(self) -> torch.optim.Optimizer:
        """Torch optimizer object which updates the synthesis target."""
        # numpydoc ignore=RT01,ES01
        return self._optimizer

    @property
    def scheduler(self) -> torch.optim.lr_scheduler.LRScheduler | None:
        """Learning rate scheduler which adjusts optimizer learning rate."""
        # numpydoc ignore=RT01,ES01
        return self._scheduler



"""
Very simple visual models.

While these may be useful as is, they are useful when combined with each other or
non-linearities, as in :mod:`plenoptic.simulate.models.frontend`
"""

from typing import Any

import torch
from torch import Tensor
from torch import nn as nn
from torch.nn import functional as F

from ...tools.conv import same_padding
from ..canonical_computations.filters import _validate_filter_args, circular_gaussian2d

__all__ = ["Identity", "Linear", "Gaussian", "CenterSurround"]


class Identity(torch.nn.Module):
    r"""
    Simple class that just returns a copy of the image.

    We use this as a "dummy model" for metrics that we don't have the
    representation for. We use this as the model and then just change
    the objective function.

    Examples
    --------
    >>> import plenoptic as po
    >>> identity_model = po.simul.Identity()
    >>> identity_model
    Identity()
    """

    def __init__(self):
        super().__init__()

    def forward(self, x: Tensor) -> Tensor:
        """
        Return a copy of the tensor.

        Parameters
        ----------
        x
            The tensor to return.

        Returns
        -------
        x
            A clone of the input tensor.

        Examples
        --------
        .. plot::

           >>> import plenoptic as po
           >>> identity_model = po.simul.Identity()
           >>> img = po.data.curie()
           >>> y = identity_model.forward(img)
           >>> titles = ["Input", "Output (identical)"]
           >>> po.imshow([img, y], title=titles)
           <PyrFigure ...>
        """  # numpydoc ignore=ES01
        y = 1 * x
        return y


class Linear(nn.Module):
    r"""
    Simplistic linear convolutional model.

    If ``default_filters=True``, this model splits the input image into low
    and high frequencies.

    Parameters
    ----------
    kernel_size
        Convolutional kernel size.
    pad_mode
        Mode with which to pad image using :func:`torch.nn.functional.pad()`.
    default_filters
        Initialize the filters to a low-pass and a band-pass. If ``False``, filters are
        randomly initialized.

    Raises
    ------
    ValueError
        If kernel_size is not one or two positive integers.

    Examples
    --------
    >>> import plenoptic as po
    >>> linear_model = po.simul.Linear()
    >>> linear_model
    Linear(
      (conv): Conv2d(1, 2, kernel_size=(3, 3), stride=(1, 1), bias=False)
    )

    To specify the kernel size :

    >>> linear_model = po.simul.Linear(kernel_size=(5, 5))
    >>> linear_model
    Linear(
      (conv): Conv2d(1, 2, kernel_size=(5, 5), stride=(1, 1), bias=False)
    )
    """

    def __init__(
        self,
        kernel_size: int | tuple[int, int] = (3, 3),
        pad_mode: str = "circular",
        default_filters: bool = True,
    ):
        super().__init__()

        self.pad_mode = pad_mode
        # std and out_channels are not used by Linear, so set to values we know will
        # pass
        self.kernel_size, _, _ = _validate_filter_args(kernel_size, 1, 1)

        self.conv = nn.Conv2d(1, 2, kernel_size, bias=False)

        if default_filters:
            var = torch.as_tensor(3.0)
            f1 = circular_gaussian2d(kernel_size, std=torch.sqrt(var))

            f2 = circular_gaussian2d(kernel_size, std=torch.sqrt(var / 3))

            f2 = f2 - f1
            f2 = f2 / f2.sum()

            self.conv.weight.data = torch.cat([f1, f2], dim=0)

    def forward(self, x: Tensor) -> Tensor:
        """
        Convolve filter with input tensor.

        We use same-padding to ensure that the output and input shapes are matched.

        Parameters
        ----------
        x
            The input tensor, with (batch, channel, height, width).

        Returns
        -------
        y
            A linear convolution of the input image, of same shape as the input.

        Examples
        --------
        .. plot::

          >>> import plenoptic as po
          >>> linear_model = po.simul.Linear()
          >>> img = po.data.curie()
          >>> y = linear_model.forward(img)
          >>> po.imshow(
          ...     [img, y],
          ...     title=[
          ...         "Input image",
          ...         "Lowpass channel output",
          ...         "Bandpass channel output",
          ...     ],
          ... )
          <PyrFigure size...>
        """
        y = same_padding(x, self.kernel_size, pad_mode=self.pad_mode)
        h = self.conv(y)
        return h


class Gaussian(nn.Module):
    """
    Isotropic Gaussian convolutional filter.

    Kernel elements are normalized and sum to one.

    Parameters
    ----------
    kernel_size
        Size of convolutional kernel.
    std
        Standard deviation of circularly symmetric Gaussian kernel.
    pad_mode
        Padding mode argument to pass to :func:`torch.nn.functional.pad`.
    out_channels
        Number of filters. If ``None``, inferred from shape of ``std``.
    cache_filt
        Whether or not to cache the filter. Avoids regenerating filt with each
        forward pass.

    Raises
    ------
    ValueError
        If out_channels is not a positive integer.
    ValueError
        If kernel_size is not a positive integer.
    ValueError
        If std is not positive.
    ValueError
        If std is non-scalar and ``len(std) != out_channels``

    Examples
    --------
    >>> import plenoptic as po
    >>> gaussian_model = po.simul.Gaussian(kernel_size=10)
    >>> gaussian_model
    Gaussian()
    """

    def __init__(
        self,
        kernel_size: int | tuple[int, int],
        std: int | list[int] | float | list[float] | Tensor = 3.0,
        pad_mode: str = "reflect",
        out_channels: int | None = None,
        cache_filt: bool = False,
    ):
        super().__init__()
        self.kernel_size, std, out_channels = _validate_filter_args(
            kernel_size, std, out_channels
        )
        self.std = nn.Parameter(std)

        self.pad_mode = pad_mode
        self.out_channels = out_channels

        self.cache_filt = cache_filt
        self.register_buffer("_filt", None)

    @property
    def filt(self) -> Tensor:
        """Gaussian filter(s)."""  # numpydoc ignore=ES01,RT01,EX01
        if self._filt is not None:  # use old filter
            return self._filt
        else:  # create new filter, optionally cache it
            filt = circular_gaussian2d(self.kernel_size, self.std, self.out_channels)

            if self.cache_filt:
                self.register_buffer("_filt", filt)
            return filt

    def forward(self, x: Tensor, **conv2d_kwargs: Any) -> Tensor:
        """
        Convolve Gaussian filter with input tensor.

        We use same-padding to ensure that the output and input shapes are matched.

        Parameters
        ----------
        x
            The input tensor, should be 4d (batch, channel, height, width).
        **conv2d_kwargs
            Passed to :func:`torch.nn.functional.conv2d`.

        Returns
        -------
        y
            A linear convolution of the input image, of same shape as the input.

        Examples
        --------
        .. plot::

          >>> import plenoptic as po
          >>> gaussian_model = po.simul.Gaussian(kernel_size=10)
          >>> img = po.data.curie()
          >>> y = gaussian_model.forward(img)
          >>> po.imshow([img, y], title=["Input image", "Output"])
          <PyrFigure size...>

        Multiple output channels with different standard deviations.

        .. plot::

          >>> import plenoptic as po
          >>> gaussian_model = po.simul.Gaussian(
          ...     kernel_size=10, std=[2, 5], out_channels=2
          ... )
          >>> img = po.data.curie()
          >>> y = gaussian_model.forward(img)
          >>> po.imshow(
          ...     [img, y],
          ...     title=["Input image", "Output Channel 0", "Output Channel 1"],
          ... )
          <PyrFigure ...>
        """
        self.std.data = self.std.data.abs()  # ensure stdev is positive

        x = same_padding(x, self.kernel_size, pad_mode=self.pad_mode)
        y = F.conv2d(x, self.filt, **conv2d_kwargs)

        return y


class CenterSurround(nn.Module):
    r"""
    Center-Surround, Difference of Gaussians (DoG) filter model.

    Can be either on-center/off-surround, or vice versa.

    Filter is constructed as:

    .. code::

        f = amplitude_ratio * center - surround
        f = f / f.sum()

    The signs of center and surround are determined by ``on_center`` argument.

    Parameters
    ----------
    kernel_size
        Shape of convolutional kernel.
    on_center
        Dictates whether center is on or off; surround will be the opposite of center
        (i.e. on-off or off-on). If List of bools, then list length must equal
        ``out_channels``, if just a single bool, then all ``out_channels`` will be
        assumed to be all on-off or off-on.
    amplitude_ratio
        Ratio of center/surround amplitude. Applied before filter normalization. Must be
        greater than or equal to 1.
    center_std
        Standard deviation of circular Gaussian for center.
    surround_std
        Standard deviation of circular Gaussian for surround.
    out_channels
        Number of filters. If ``None``, inferred from shape of ``center_std``.
    pad_mode
        Padding for convolution.
    cache_filt
        Whether or not to cache the filter. Avoids regenerating filt with each
        forward pass.

    Raises
    ------
    ValueError
        If out_channels is not a positive integer.
    ValueError
        If kernel_size is not a positive integer.
    ValueError
        If center_std or surround_std are not positive.
    ValueError
        If center_std and surround_std do not have the same number of values.
    ValueError
        If center_std or surround_std are non-scalar and their lengths do not
        equal ``out_channels``

    Examples
    --------
    >>> import plenoptic as po
    >>> cs_model = po.simul.CenterSurround(kernel_size=10)
    >>> cs_model
    CenterSurround()

    Model with both on-center/off-surround and off-center/on-surround:

    >>> import plenoptic as po
    >>> cs_model = po.simul.CenterSurround(10, [True, False])
    >>> cs_model
    CenterSurround()
    """

    def __init__(
        self,
        kernel_size: int | tuple[int, int],
        on_center: bool | list[bool] = True,
        amplitude_ratio: float = 1.25,
        center_std: int | list[int] | float | list[float] | Tensor = 1.0,
        surround_std: int | list[int] | float | list[float] | Tensor = 4.0,
        out_channels: int | None = None,
        pad_mode: str = "reflect",
        cache_filt: bool = False,
    ):
        super().__init__()

        on_center = torch.as_tensor(on_center)
        if out_channels is None and on_center.numel() != 1:
            out_channels = len(on_center)

        self.kernel_size, center_std, out_channels = _validate_filter_args(
            kernel_size,
            center_std,
            out_channels,
            "center_std",
        )
        _, surround_std, _ = _validate_filter_args(
            kernel_size, surround_std, out_channels, "surround_std", "len(center_std)"
        )

        self.center_std = nn.Parameter(center_std)
        self.surround_std = nn.Parameter(surround_std)

        # make sure each channel is on-off or off-on
        if on_center.numel() == 1:
            on_center = on_center.repeat(out_channels)
        if len(on_center) != out_channels:
            raise ValueError("len(on_center) must equal out_channels")
        self.on_center = on_center

        amplitude_ratio = torch.as_tensor(amplitude_ratio)
        if amplitude_ratio.nelement() > 1:
            raise ValueError("amplitude_ratio must be a scalar")
        if amplitude_ratio < 1.0:
            raise ValueError("amplitude_ratio must at least be 1.")
        self.register_buffer("amplitude_ratio", amplitude_ratio)

        self.out_channels = out_channels
        self.pad_mode = pad_mode

        self.cache_filt = cache_filt
        self.register_buffer("_filt", None)

    @property
    def filt(self) -> Tensor:
        """Center-surround filter(s)."""  # numpydoc ignore=ES01,RT01,EX01
        if self._filt is not None:
            # use cached filt
            return self._filt
        else:
            # generate new filt and optionally cache

            filt_center = circular_gaussian2d(
                self.kernel_size, self.center_std, self.out_channels
            )
            filt_surround = circular_gaussian2d(
                self.kernel_size, self.surround_std, self.out_channels
            )

            # sign is + or - depending on center is on or off
            sign = torch.as_tensor(
                [1.0 if x else -1.0 for x in self.on_center],
                device=self.amplitude_ratio.device,
            )
            sign = sign.view(self.out_channels, 1, 1, 1)
            filt = self.amplitude_ratio * (sign * (filt_center - filt_surround))

            if self.cache_filt:
                self.register_buffer("_filt", filt)
        return filt

    def forward(self, x: Tensor) -> Tensor:
        """
        Convolve center-surround filter with input tensor.

        We use same-padding to ensure that the output and input shapes are matched.

        Parameters
        ----------
        x
            The input tensor, should be 4d (batch, channel, height, width).

        Returns
        -------
        y
            A linear convolution of the input image, of same shape as the input.

        Examples
        --------
        .. plot::

          >>> import plenoptic as po
          >>> cs_model = po.simul.CenterSurround(kernel_size=10)
          >>> img = po.data.curie()
          >>> y = cs_model.forward(img)
          >>> po.imshow([img, y], title=["Input image", "Output"])
          <PyrFigure size...>

        Model with both on-center/off-surround and off-center/on-surround:

        .. plot::

          >>> import plenoptic as po
          >>> cs_model = po.simul.CenterSurround(10, [True, False])
          >>> img = po.data.curie()
          >>> y = cs_model.forward(img)
          >>> titles = [
          ...     "Input image",
          ...     "On-center/off-surround",
          ...     "Off-center/on-surround",
          ... ]
          >>> po.imshow([img, y], title=titles)
          <PyrFigure size...>
        """
        x = same_padding(x, self.kernel_size, pad_mode=self.pad_mode)
        y = F.conv2d(x, self.filt, bias=None)
        return y



import math
import os.path as op
from contextlib import nullcontext as does_not_raise

import pytest
import torch

import plenoptic as po
from conftest import DEVICE, check_loss_saved_synth


# in order for pickling to work with functions, they must be defined at top of
# module: https://stackoverflow.com/a/36995008
def custom_loss(x1, x2):
    return (x1 - x2).sum()


class TestMetamers:
    @pytest.mark.parametrize(
        "model", ["frontend.LinearNonlinear.nograd"], indirect=True
    )
    @pytest.mark.parametrize("loss_func", ["mse", "l2", "custom"])
    @pytest.mark.parametrize(
        "fail",
        [False, "img", "model", "loss", "range_penalty", "dtype", "allowed_range"],
    )
    @pytest.mark.parametrize("range_penalty", [0.1, 0])
    @pytest.mark.parametrize("allowed_range", [(0, 1), (-1, 1)])
    def test_save_load(
        self,
        einstein_img,
        model,
        loss_func,
        fail,
        range_penalty,
        allowed_range,
        tmp_path,
    ):
        if loss_func == "mse":
            loss = po.tools.optim.mse
        elif loss_func == "l2":
            loss = po.tools.optim.l2_norm
        elif loss_func == "custom":
            loss = custom_loss
        met = po.synth.Metamer(
            einstein_img,
            model,
            loss_function=loss,
            allowed_range=allowed_range,
            range_penalty_lambda=range_penalty,
        )
        met.synthesize(max_iter=4, store_progress=True)
        met.save(op.join(tmp_path, "test_metamer_save_load.pt"))
        if fail:
            if fail == "img":
                einstein_img = torch.rand_like(einstein_img)
                expectation = pytest.raises(
                    ValueError,
                    match="Saved and initialized attribute image have different values",
                )
            elif fail == "model":
                model = po.simul.Gaussian(30).to(DEVICE)
                po.tools.remove_grad(model)
                model.eval()
                expectation = pytest.raises(
                    ValueError,
                    match=("Saved and initialized model output have different values"),
                )
            elif fail == "loss":
                loss = po.tools.optim.relative_sse
                expectation = pytest.raises(
                    ValueError,
                    match=(
                        "Saved and initialized loss_function output have different "
                        "values"
                    ),
                )
            elif fail == "allowed_range":
                allowed_range = (0, 5)
                expectation = pytest.raises(
                    ValueError,
                    match=("Saved and initialized allowed_range are different"),
                )
            elif fail == "range_penalty":
                range_penalty = 0.5
                expectation = pytest.raises(
                    ValueError,
                    match=("Saved and initialized range_penalty_lambda are different"),
                )
            elif fail == "dtype":
                einstein_img = einstein_img.to(torch.float64)
                # need to create new instance of model, because otherwise the
                # version with doubles as weights will persist for other tests
                model = po.simul.LinearNonlinear((31, 31)).to(DEVICE)
                po.tools.remove_grad(model)
                model.to(torch.float64)
                model.eval()
                expectation = pytest.raises(
                    ValueError,
                    match="Saved and initialized attribute image have different dtype",
                )
            met_copy = po.synth.Metamer(
                einstein_img,
                model,
                loss_function=loss,
                allowed_range=allowed_range,
                range_penalty_lambda=range_penalty,
            )
            with expectation:
                met_copy.load(
                    op.join(tmp_path, "test_metamer_save_load.pt"),
                    map_location=DEVICE,
                )
        else:
            met_copy = po.synth.Metamer(
                einstein_img,
                model,
                loss_function=loss,
                range_penalty_lambda=range_penalty,
                allowed_range=allowed_range,
            )
            met_copy.load(
                op.join(tmp_path, "test_metamer_save_load.pt"),
                map_location=DEVICE,
            )
            for k in [
                "image",
                "saved_metamer",
                "metamer",
                "target_representation",
            ]:
                if not getattr(met, k).allclose(getattr(met_copy, k), rtol=1e-2):
                    raise ValueError(
                        f"Something went wrong with saving and loading! {k} not"
                        " the same"
                    )
            # check loss functions correctly saved
            met_loss = met.loss_function(
                met.model(met.metamer), met.target_representation
            )
            met_copy_loss = met_copy.loss_function(
                met.model(met.metamer), met_copy.target_representation
            )
            if not torch.allclose(met_loss, met_copy_loss, rtol=1e-2):
                raise ValueError(
                    "Loss function not properly saved! Before saving was"
                    f" {met_loss}, after loading was {met_copy_loss}"
                )
            # check that can resume
            met_copy.synthesize(
                max_iter=4,
                store_progress=True,
            )

    @pytest.mark.parametrize(
        "model", ["frontend.LinearNonlinear.nograd"], indirect=True
    )
    def test_setup_initial_image(self, einstein_img, curie_img, model):
        met = po.synth.Metamer(einstein_img, model)
        met.setup(curie_img)
        met.synthesize(5)

    @pytest.mark.parametrize(
        "model", ["frontend.LinearNonlinear.nograd"], indirect=True
    )
    @pytest.mark.parametrize("warn", ["shape", "ndim"])
    @pytest.mark.filterwarnings("ignore:.*mostly been tested on 4d inputs:UserWarning")
    def test_setup_initial_image_warn(self, einstein_img, curie_img, model, warn):
        met = po.synth.Metamer(einstein_img, model)
        if warn == "shape":
            img = curie_img[..., :255, :255]
        elif warn == "ndim":
            img = curie_img[0]
        txt = "initial_image and image are different sizes"
        with pytest.warns(UserWarning, match=txt):
            met.setup(img)

    @pytest.mark.parametrize(
        "model", ["frontend.LinearNonlinear.nograd"], indirect=True
    )
    def test_setup_fail(self, einstein_img, model):
        met = po.synth.Metamer(einstein_img, model)
        met.setup()
        with pytest.raises(ValueError, match=r"setup\(\) can only be called once"):
            met.setup()

    @pytest.mark.parametrize(
        "model", ["frontend.LinearNonlinear.nograd"], indirect=True
    )
    @pytest.mark.filterwarnings("ignore:You will need to call setup:UserWarning")
    def test_synth_then_setup(self, einstein_img, model, tmp_path):
        met = po.synth.Metamer(einstein_img, model)
        met.setup(optimizer=torch.optim.SGD)
        met.synthesize(max_iter=4)
        met.save(op.join(tmp_path, "test_metamer_synth_then_setup.pt"))
        met = po.synth.Metamer(einstein_img, model)
        met.load(op.join(tmp_path, "test_metamer_synth_then_setup.pt"))
        with pytest.raises(ValueError, match="Don't know how to initialize"):
            met.synthesize(5)
        met.setup(optimizer=torch.optim.SGD)
        met.synthesize(5)

    @pytest.mark.parametrize(
        "model", ["frontend.LinearNonlinear.nograd"], indirect=True
    )
    def test_setup_load_fail(self, einstein_img, model, tmp_path):
        met = po.synth.Metamer(einstein_img, model)
        met.synthesize(max_iter=4)
        met.save(op.join(tmp_path, "test_metamer_setup_load_fail.pt"))
        met = po.synth.Metamer(einstein_img, model)
        met.load(op.join(tmp_path, "test_metamer_setup_load_fail.pt"))
        with pytest.raises(
            ValueError, match="Cannot set initial_image after calling load"
        ):
            met.setup(po.data.curie())

    @pytest.mark.parametrize(
        "model", ["frontend.LinearNonlinear.nograd"], indirect=True
    )
    @pytest.mark.parametrize("fail", ["synth", "setup", "continue"])
    def test_load_init_fail(self, einstein_img, model, fail, tmp_path):
        met = po.synth.Metamer(einstein_img, model)
        met.synthesize(max_iter=4, store_progress=True)
        met.save(op.join(tmp_path, "test_metamer_load_init_fail.pt"))
        if fail != "continue":
            met = po.synth.Metamer(einstein_img, model)
            if fail == "setup":
                met.setup()
            elif fail == "synth":
                met.synthesize(max_iter=4, store_progress=True)
        with pytest.raises(
            ValueError, match="load can only be called with a just-initialized"
        ):
            met.load(op.join(tmp_path, "test_metamer_load_init_fail.pt"))

    @pytest.mark.parametrize(
        "model", ["frontend.LinearNonlinear.nograd"], indirect=True
    )
    @pytest.mark.parametrize("fail", [False, "name", "behavior"])
    def test_load_names(self, fail, einstein_img, model, tmp_path):
        # name and behavior same as our LinearNonlinear, but module path is
        # different
        if fail is False:

            class LinearNonlinear(torch.nn.Module):
                def __init__(self, *args, **kwargs):
                    super().__init__()
                    self.model = po.simul.LinearNonlinear((31, 31)).to(DEVICE)

                def forward(self, *args, **kwargs):
                    return self.model(*args, **kwargs)

            model2 = LinearNonlinear()
            expectation = does_not_raise()
        # name different but behavior same
        elif fail == "name":

            class LinearNonlinearFAIL(torch.nn.Module):
                def __init__(self, *args, **kwargs):
                    super().__init__()
                    self.model = po.simul.LinearNonlinear((31, 31)).to(DEVICE)

                def forward(self, *args, **kwargs):
                    return self.model(*args, **kwargs)

            model2 = LinearNonlinearFAIL()
            expectation = pytest.raises(
                ValueError, match="Saved and initialized model have different names"
            )
        # name same but behavior different
        elif fail == "behavior":

            class LinearNonlinear(torch.nn.Module):
                def __init__(self, *args, **kwargs):
                    super().__init__()
                    self.model = po.simul.LinearNonlinear((16, 16)).to(DEVICE)

                def forward(self, *args, **kwargs):
                    return self.model(*args, **kwargs)

            model2 = LinearNonlinear()
            expectation = pytest.raises(
                ValueError,
                match="Saved and initialized model output have different values",
            )
        met = po.synth.Metamer(einstein_img, model)
        met.synthesize(max_iter=4, store_progress=True)
        met.save(op.join(tmp_path, f"test_metamer_load_names_{fail}.pt"))
        po.tools.remove_grad(model2)
        model2.eval()
        met = po.synth.Metamer(einstein_img, model2)
        with expectation:
            met.load(op.join(tmp_path, f"test_metamer_load_names_{fail}.pt"))

    @pytest.mark.parametrize(
        "model", ["frontend.LinearNonlinear.nograd"], indirect=True
    )
    def test_examine_saved_object(self, einstein_img, model, tmp_path):
        met = po.synth.Metamer(einstein_img, model)
        met.synthesize(max_iter=4, store_progress=True)
        met.save(op.join(tmp_path, "test_metamer_examine.pt"))
        po.tools.examine_saved_synthesis(op.join(tmp_path, "test_metamer_examine.pt"))

    @pytest.mark.parametrize(
        "model", ["frontend.LinearNonlinear.nograd"], indirect=True
    )
    @pytest.mark.parametrize("synth_type", ["eig", "mad"])
    def test_load_object_type(self, einstein_img, model, synth_type, tmp_path):
        met = po.synth.Metamer(einstein_img, model)
        met.synthesize(max_iter=4, store_progress=True)
        met.save(op.join(tmp_path, "test_metamer_load_object_type.pt"))
        if synth_type == "eig":
            met = po.synth.Eigendistortion(einstein_img, model)
        elif synth_type == "mad":
            met = po.synth.MADCompetition(
                einstein_img,
                po.metric.mse,
                po.metric.mse,
                "min",
                metric_tradeoff_lambda=1,
            )
        with pytest.raises(
            ValueError, match="Saved object was a.* but initialized object is"
        ):
            met.load(op.join(tmp_path, "test_metamer_load_object_type.pt"))

    @pytest.mark.parametrize(
        "model", ["frontend.LinearNonlinear.nograd"], indirect=True
    )
    @pytest.mark.parametrize("model_behav", ["dtype", "shape", "name"])
    def test_load_model_change(self, einstein_img, model, model_behav, tmp_path):
        met = po.synth.Metamer(einstein_img, model)
        met.synthesize(max_iter=4, store_progress=True)
        met.save(op.join(tmp_path, "test_metamer_load_model_change.pt"))
        if model_behav == "dtype":
            # this actually gets raised in the model validation step (during init), not
            # load.
            expectation = pytest.raises(TypeError, match="model changes precision")
        elif model_behav == "shape":
            expectation = pytest.raises(
                ValueError,
                match="Saved and initialized model output have different shape",
            )
        elif model_behav == "name":
            expectation = pytest.raises(
                ValueError, match="Saved and initialized model have different names"
            )

        class NewModel(torch.nn.Module):
            def __init__(self):
                super().__init__()
                self.model = model

            def forward(self, x):
                if model_behav == "dtype":
                    return self.model(x).to(torch.float64)
                elif model_behav == "shape":
                    return self.model(x).flatten(-2)
                elif model_behav == "name":
                    return self.model(x)

        model = NewModel()
        model.eval()
        with expectation:
            met = po.synth.Metamer(einstein_img, model)
            met.load(op.join(tmp_path, "test_metamer_load_model_change.pt"))

    @pytest.mark.parametrize(
        "model", ["frontend.LinearNonlinear.nograd"], indirect=True
    )
    @pytest.mark.parametrize("loss_behav", ["dtype", "shape", "name"])
    def test_load_loss_change(self, einstein_img, model, loss_behav, tmp_path):
        met = po.synth.Metamer(einstein_img, model)
        met.synthesize(max_iter=4, store_progress=True)
        met.save(op.join(tmp_path, "test_metamer_load_loss_change.pt"))

        def new_loss(x, y):
            if loss_behav == "dtype":
                return po.tools.optim.mse(x, y).to(torch.float64)
            elif loss_behav == "shape":
                return torch.stack(
                    [po.tools.optim.mse(x, y) for _ in range(2)]
                ).unsqueeze(0)
            elif loss_behav == "name":
                return po.tools.optim.mse(x, y)

        met = po.synth.Metamer(einstein_img, model, loss_function=new_loss)
        if loss_behav == "name":
            expectation_str = "Saved and initialized loss_function have different names"
        else:
            expectation_str = (
                "Saved and initialized loss_function output have different"
                f" {loss_behav}"
            )
        with pytest.raises(ValueError, match=expectation_str):
            met.load(op.join(tmp_path, "test_metamer_load_loss_change.pt"))

    @pytest.mark.parametrize(
        "model", ["frontend.LinearNonlinear.nograd"], indirect=True
    )
    @pytest.mark.parametrize("attribute", ["saved", "init"])
    def test_load_attributes(self, einstein_img, model, attribute, tmp_path):
        met = po.synth.Metamer(einstein_img, model)
        met.synthesize(max_iter=4, store_progress=True)
        if attribute == "saved":
            met.test = "BAD"
            err_str = "Saved"
        met.save(op.join(tmp_path, "test_metamer_load_attributes.pt"))
        met = po.synth.Metamer(einstein_img, model)
        if attribute == "init":
            met.test = "BAD"
            err_str = "Initialized"
        with pytest.raises(
            ValueError, match=rf"{err_str} object has 1 attribute\(s\) not present"
        ):
            met.load(op.join(tmp_path, "test_metamer_load_attributes.pt"))

    @pytest.mark.parametrize(
        "model",
        ["frontend.OnOff.nograd.ctf"],
        indirect=True,
    )
    @pytest.mark.parametrize(
        "optim_opts",
        [
            None,
            "SGD",
            "SGD-args",
            "Adam",
            "Adam-args",
            "Scheduler",
            "Scheduler-args",
            "LBFGS",
            "LBFGS-args",
        ],
    )
    @pytest.mark.parametrize("fail", [True, False])
    @pytest.mark.parametrize("coarse_to_fine", [True, False])
    @pytest.mark.filterwarnings("ignore:You will need to call setup:UserWarning")
    @pytest.mark.filterwarnings(
        "ignore:Validating whether model can work with coarse-to-fine:UserWarning"
    )
    def test_load_optimizer(
        self, curie_img, model, coarse_to_fine, optim_opts, fail, tmp_path
    ):
        if coarse_to_fine:
            met = po.synth.MetamerCTF(curie_img, model)
        else:
            met = po.synth.Metamer(curie_img, model)
        scheduler = None
        optimizer = None
        optimizer_kwargs = None
        scheduler_kwargs = None
        check_optimizer = [torch.optim.Adam, {"eps": 1e-8, "lr": 0.01}]
        check_scheduler = None
        if optim_opts is not None:
            if "Scheduler" in optim_opts:
                scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau
                check_scheduler = [
                    torch.optim.lr_scheduler.ReduceLROnPlateau,
                    {"factor": 0.1},
                ]
                if "args" in optim_opts:
                    scheduler_kwargs = {"factor": 1e-3}
                    check_scheduler[1] = {"factor": 1e-3}
            else:
                if "Adam" in optim_opts:
                    optimizer = torch.optim.Adam
                elif "SGD" in optim_opts:
                    optimizer = torch.optim.SGD
                    check_optimizer[0] = torch.optim.SGD
                    check_optimizer[1] = {"lr": 0.01}
                elif "LBFGS" in optim_opts:
                    optimizer = torch.optim.LBFGS
                    check_optimizer[0] = torch.optim.LBFGS
                    check_optimizer[1] = {"lr": 0.01}
                if "args" in optim_opts:
                    optimizer_kwargs = {"lr": 1}
                    check_optimizer[1] = {"lr": 1}
        met.setup(
            optimizer=optimizer,
            scheduler=scheduler,
            optimizer_kwargs=optimizer_kwargs,
            scheduler_kwargs=scheduler_kwargs,
        )
        met.synthesize(max_iter=5)
        met.save(op.join(tmp_path, "test_metamer_optimizer.pt"))
        if coarse_to_fine:
            met = po.synth.MetamerCTF(curie_img, model)
        else:
            met = po.synth.Metamer(curie_img, model)
        met.load(op.join(tmp_path, "test_metamer_optimizer.pt"))
        optimizer_kwargs = None
        scheduler_kwargs = None
        if not fail:
            if optim_opts is not None:
                if "Adam" in optim_opts:
                    optimizer = torch.optim.Adam
                elif "SGD" in optim_opts:
                    optimizer = torch.optim.SGD
                elif "LBFGS" in optim_opts:
                    optimizer = torch.optim.LBFGS
                if "Scheduler" in optim_opts:
                    scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau
            expectation = does_not_raise()
        else:
            expect_str = "User-specified optimizer must have same type"
            if optim_opts is None:
                optimizer = torch.optim.SGD
            else:
                if optim_opts == "Adam":
                    optimizer = torch.optim.SGD
                elif optim_opts == "Adam-args":
                    optimizer = torch.optim.Adam
                    optimizer_kwargs = {"lr": 1}
                    expect_str = (
                        "When initializing optimizer after load, optimizer_kwargs"
                    )
                elif optim_opts == "SGD":
                    optimizer = None
                    expect_str = "Don't know how to initialize saved optimizer"
                elif optim_opts == "SGD-args":
                    optimizer = torch.optim.SGD
                    optimizer_kwargs = {"lr": 1}
                    expect_str = (
                        "When initializing optimizer after load, optimizer_kwargs"
                    )
                elif optim_opts == "LBFGS":
                    optimizer = None
                    expect_str = "Don't know how to initialize saved optimizer"
                elif optim_opts == "LBFGS-args":
                    optimizer = torch.optim.LBFGS
                    optimizer_kwargs = {"lr": 1}
                    expect_str = (
                        "When initializing optimizer after load, optimizer_kwargs"
                    )
                elif optim_opts == "Scheduler":
                    scheduler = torch.optim.lr_scheduler.ConstantLR
                    expect_str = "User-specified scheduler must have same type"
                elif optim_opts == "Scheduler-args":
                    scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau
                    scheduler_kwargs = {"factor": 1e-3}
                    expect_str = (
                        "When initializing scheduler after load, scheduler_kwargs"
                    )
            expectation = pytest.raises(ValueError, match=expect_str)
        # these fail during setup
        with expectation:
            met.setup(
                optimizer=optimizer,
                scheduler=scheduler,
                optimizer_kwargs=optimizer_kwargs,
                scheduler_kwargs=scheduler_kwargs,
            )
            met.synthesize(max_iter=5)
            if not isinstance(met.optimizer, check_optimizer[0]):
                raise ValueError("Didn't properly set optimizer!")
            state_dict = met.optimizer.state_dict()["param_groups"][0]
            for k, v in check_optimizer[1].items():
                if state_dict[k] != v:
                    raise ValueError(
                        "Didn't properly set optimizer kwargs! "
                        f"Expected {v} but got {state_dict[k]}!"
                    )
            if check_scheduler is not None:
                if not isinstance(met.scheduler, check_scheduler[0]):
                    raise ValueError("Didn't properly set scheduler!")
                state_dict = met.scheduler.state_dict()
                for k, v in check_scheduler[1].items():
                    if met.scheduler.state_dict()[k] != v:
                        raise ValueError("Didn't properly set scheduler kwargs!")
            elif met.scheduler is not None:
                raise ValueError("Didn't set scheduler to None!")

    @pytest.mark.parametrize(
        "model", ["frontend.LinearNonlinear.nograd"], indirect=True
    )
    def test_load_tol(self, einstein_img, model, tmp_path):
        met = po.synth.Metamer(einstein_img, model, allowed_range=(-1, 2))
        met.synthesize(5)
        met.save(op.join(tmp_path, "test_metamer_load_tol.pt"))
        met = po.synth.Metamer(
            einstein_img + 1e-7 * torch.rand_like(einstein_img),
            model,
            allowed_range=(-1, 2),
        )
        with pytest.raises(ValueError, match="Saved and initialized attribute image"):
            met.load(op.join(tmp_path, "test_metamer_load_tol.pt"))
        met.load(
            op.join(tmp_path, "test_metamer_load_tol.pt"), tensor_equality_atol=1e-7
        )

    @pytest.mark.parametrize(
        "model", ["frontend.LinearNonlinear.nograd"], indirect=True
    )
    @pytest.mark.parametrize("load", [True, False])
    @pytest.mark.filterwarnings("ignore:You will need to call setup:UserWarning")
    def test_resume_synthesis(self, einstein_img, curie_img, model, load, tmp_path):
        met = po.synth.Metamer(einstein_img, model)
        # Adam has some stochasticity in its initialization(?), so this test doesn't
        # quite work with it (it does if you do po.tools.set_seed(2) at the top of the
        # function)
        met.setup(curie_img, optimizer=torch.optim.SGD)
        met.synthesize(10)
        met_copy = po.synth.Metamer(einstein_img, model)
        met_copy.setup(curie_img, optimizer=torch.optim.SGD)
        met_copy.synthesize(5)
        if load:
            met_copy.save(op.join(tmp_path, "test_metamer_resume_synthesis.pt"))
            met_copy = po.synth.Metamer(einstein_img, model)
            met_copy.load(op.join(tmp_path, "test_metamer_resume_synthesis.pt"))
            met_copy.setup(optimizer=torch.optim.SGD)
            met_copy.synthesize(5)
        else:
            met_copy.synthesize(5)
        if not torch.allclose(met.metamer, met_copy.metamer):
            raise ValueError("Resuming synthesis different than just continuing!")

    # test that we support models with 3d and 4d outputs
    @pytest.mark.parametrize(
        "model",
        ["PortillaSimoncelli", "frontend.LinearNonlinear.nograd"],
        indirect=True,
    )
    def test_model_dimensionality_real(self, einstein_img, model):
        met = po.synth.Metamer(einstein_img, model)
        met.synthesize(5)

    @pytest.mark.parametrize(
        "model",
        [f"diff_dims-{i}" for i in range(1, 6)],
        indirect=True,
    )
    @pytest.mark.parametrize("input_dim", [2, 3, 4, 5])
    @pytest.mark.filterwarnings("ignore:.*mostly been tested on 4d inputs:UserWarning")
    @pytest.mark.filterwarnings(
        "ignore:.*mostly been tested on models which:UserWarning"
    )
    def test_dimensionality(self, einstein_img, input_dim, model):
        img = einstein_img.squeeze()
        while img.ndimension() < input_dim:
            img = img.unsqueeze(0)
        met = po.synth.Metamer(img, model)
        met.synthesize(5)

    @pytest.mark.parametrize(
        "model",
        ["frontend.LinearNonlinear.nograd", "PortillaSimoncelli"],
        indirect=True,
    )
    @pytest.mark.parametrize("store_progress", [True, 2, 3])
    @pytest.mark.filterwarnings(
        "ignore:Validating whether model can work with coarse-to-fine:UserWarning"
    )
    def test_store_rep(self, einstein_img, model, store_progress):
        if hasattr(model, "scales"):
            metamer = po.synth.MetamerCTF(einstein_img, model)
        else:
            metamer = po.synth.Metamer(einstein_img, model)
        max_iter = 3
        if store_progress == 3:
            max_iter = 6
        metamer.synthesize(max_iter=max_iter, store_progress=store_progress)
        check_loss_saved_synth(
            metamer.losses,
            metamer.saved_metamer,
            max_iter,
            metamer.objective_function,
            metamer.store_progress,
        )
        metamer.synthesize(max_iter=max_iter, store_progress=store_progress)
        check_loss_saved_synth(
            metamer.losses,
            metamer.saved_metamer,
            2 * max_iter,
            metamer.objective_function,
            metamer.store_progress,
        )

    @pytest.mark.parametrize(
        "model",
        ["frontend.LinearNonlinear.nograd", "PortillaSimoncelli"],
        indirect=True,
    )
    @pytest.mark.filterwarnings(
        "ignore:Validating whether model can work with coarse-to-fine:UserWarning"
    )
    def test_save_metamer_empty(self, einstein_img, model):
        if hasattr(model, "scales"):
            metamer = po.synth.MetamerCTF(einstein_img, model)
        else:
            metamer = po.synth.Metamer(einstein_img, model)
        torch.equal(metamer.saved_metamer, torch.empty(0))
        metamer.synthesize(max_iter=3)
        torch.equal(metamer.saved_metamer, metamer.metamer.to("cpu"))

    @pytest.mark.parametrize(
        "model",
        ["frontend.LinearNonlinear.nograd", "PortillaSimoncelli"],
        indirect=True,
    )
    @pytest.mark.filterwarnings(
        "ignore:Validating whether model can work with coarse-to-fine:UserWarning"
    )
    def test_metamer_empty_loss(self, einstein_img, model):
        if hasattr(model, "scales"):
            met = po.synth.MetamerCTF(einstein_img, model)
        else:
            met = po.synth.Metamer(einstein_img, model)
        assert met.objective_function().numel() == 0
        torch.equal(met.losses, torch.empty(0))
        met.setup()
        assert isinstance(met.objective_function(), torch.Tensor)
        assert met.losses.numel() > 0
        met.synthesize(max_iter=2)
        assert isinstance(met.objective_function(), torch.Tensor)
        assert met.losses.numel() > 0

    @pytest.mark.parametrize("iteration", [None, 0, -2, -3, 2, 1, 6, -7])
    @pytest.mark.parametrize("store_progress", [True, False, 2])
    @pytest.mark.parametrize("iteration_selection", ["floor", "ceiling", "round"])
    @pytest.mark.parametrize(
        "model",
        ["frontend.LinearNonlinear.nograd", "PortillaSimoncelli"],
        indirect=True,
    )
    @pytest.mark.filterwarnings(
        "ignore:Validating whether model can work with coarse-to-fine:UserWarning"
    )
    def test_metamer_get_progress(
        self, einstein_img, model, iteration, store_progress, iteration_selection
    ):
        if hasattr(model, "scales"):
            met = po.synth.MetamerCTF(einstein_img, model)
        else:
            met = po.synth.Metamer(einstein_img, model)
        met.synthesize(max_iter=5, store_progress=store_progress)
        if iteration_selection == "floor":
            func = math.floor
        elif iteration_selection == "ceiling":
            func = math.ceil
        else:
            func = round
        expected_dict = {}
        if iteration is None:
            expected_dict["losses"] = met.losses[-1]
            expected_dict.update(
                {
                    "iteration": 5,
                    "gradient_norm": None,
                    "pixel_change_norm": None,
                }
            )
            if store_progress:
                expected_dict.update(
                    {
                        "saved_metamer": met.metamer.to("cpu"),
                        "store_progress_iteration": 5,
                    }
                )
        elif iteration not in [6, -7]:
            expected_dict["losses"] = met.losses[iteration]
            if iteration < 0:
                # add one to account for loss and these attributes being off by one
                expected_dict.update(
                    {
                        "iteration": 6 + iteration,
                        "gradient_norm": met.gradient_norm[iteration + 1],
                        "pixel_change_norm": met.pixel_change_norm[iteration + 1],
                    }
                )
                if store_progress:
                    iter = func((6 + iteration) / store_progress)
                    expected_dict.update(
                        {
                            "saved_metamer": met.saved_metamer[iter],
                            "store_progress_iteration": iter * store_progress,
                        }
                    )
            else:
                expected_dict.update(
                    {
                        "iteration": iteration,
                        "gradient_norm": met.gradient_norm[iteration],
                        "pixel_change_norm": met.pixel_change_norm[iteration],
                    }
                )
                if store_progress:
                    iter = func(iteration / store_progress)
                    expected_dict.update(
                        {
                            "saved_metamer": met.saved_metamer[iter],
                            "store_progress_iteration": iter * store_progress,
                        }
                    )
        if iteration in [6, -7]:
            expectation = pytest.raises(IndexError, match=".*out of bounds with.*")
        elif store_progress == 2 and iteration in [1, -3]:
            expectation = pytest.warns(
                UserWarning, match="loss iteration and iteration"
            )
        else:
            expectation = does_not_raise()
        with expectation:
            progress = met.get_progress(iteration, iteration_selection)
            assert progress.keys() == expected_dict.keys()
            for k, v in progress.items():
                if isinstance(v, torch.Tensor):
                    assert torch.equal(v, expected_dict[k]), f"{k} not as expected!"
                else:
                    assert v == expected_dict[k], f"{k} not as expected!"

    @pytest.mark.parametrize(
        "model", ["frontend.LinearNonlinear.nograd"], indirect=True
    )
    def test_continue(self, einstein_img, model):
        metamer = po.synth.Metamer(einstein_img, model)
        metamer.synthesize(max_iter=3, store_progress=True)
        metamer.synthesize(max_iter=3, store_progress=True)

    @pytest.mark.parametrize("model", ["SPyr"], indirect=True)
    @pytest.mark.parametrize("coarse_to_fine", ["separate", "together"])
    @pytest.mark.filterwarnings(
        "ignore:Validating whether model can work with coarse-to-fine:UserWarning"
    )
    @pytest.mark.filterwarnings("ignore:Loss has converged:UserWarning")
    def test_coarse_to_fine(self, einstein_img, model, coarse_to_fine, tmp_path):
        metamer = po.synth.MetamerCTF(
            einstein_img, model, coarse_to_fine=coarse_to_fine
        )
        metamer.synthesize(
            max_iter=5,
            stop_iters_to_check=1,
            change_scale_criterion=10,
            ctf_iters_to_check=1,
        )
        assert len(metamer.scales_finished) > 0, "Didn't actually switch scales!"

        metamer.save(op.join(tmp_path, "test_metamer_ctf.pt"))
        metamer_copy = po.synth.MetamerCTF(
            einstein_img, model, coarse_to_fine=coarse_to_fine
        )
        metamer_copy.load(op.join(tmp_path, "test_metamer_ctf.pt"), map_location=DEVICE)
        # check the ctf-related attributes all saved correctly
        for k in [
            "coarse_to_fine",
            "scales",
            "scales_loss",
            "scales_timing",
            "scales_finished",
        ]:
            if not getattr(metamer, k) == (getattr(metamer_copy, k)):
                raise ValueError(
                    f"Something went wrong with saving and loading! {k} not the same"
                )
        # check we can resume
        metamer.synthesize(
            max_iter=5,
            stop_iters_to_check=1,
            change_scale_criterion=10,
            ctf_iters_to_check=1,
        )

    @pytest.mark.parametrize("model", ["frontend.OnOff.nograd.ctf"], indirect=True)
    @pytest.mark.parametrize(
        "optimizer",
        [
            "SGD",
            "SGD-args",
            "Adam",
            "Adam-args",
            None,
            "Scheduler-args",
            "Scheduler",
            "LBFGS",
            "LBFGS-args",
        ],
    )
    @pytest.mark.parametrize("coarse_to_fine", [True, False])
    @pytest.mark.filterwarnings(
        "ignore:Validating whether model can work with coarse-to-fine:UserWarning"
    )
    def test_optimizer(self, curie_img, model, coarse_to_fine, optimizer):
        if coarse_to_fine:
            met = po.synth.MetamerCTF(curie_img, model)
        else:
            met = po.synth.Metamer(curie_img, model)
        optimizer = None
        scheduler = None
        optimizer_kwargs = None
        scheduler_kwargs = None
        check_optimizer = [torch.optim.Adam, {"eps": 1e-8, "lr": 0.01}]
        check_scheduler = None
        if optimizer == "Adam":
            optimizer = torch.optim.Adam
        elif optimizer == "Adam-args":
            optimizer = torch.optim.Adam
            optimizer_kwargs = {"eps": 1e-5}
            check_optimizer[1]["eps"] = 1e-5
        elif optimizer == "SGD":
            optimizer = torch.optim.SGD
            check_optimizer = [torch.optim.SGD, {"lr": 0.01}]
        elif optimizer == "SGD-args":
            optimizer = torch.optim.SGD
            optimizer_kwargs = {"lr": 1}
            check_optimizer = [torch.optim.SGD, {"lr": 1}]
        elif optimizer == "LBFGS":
            optimizer = torch.optim.LBFGS
            check_optimizer = [torch.optim.LBFGS, {"lr": 0.01}]
        elif optimizer == "LBFGS-args":
            optimizer = torch.optim.LBFGS
            optimizer_kwargs = {"lr": 1, "history_size": 10}
            check_optimizer = [torch.optim.LBFGS, {"lr": 1, "history_size": 10}]
        elif optimizer == "Scheduler":
            scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau
            check_scheduler = [
                torch.optim.lr_scheduler.ReduceLROnPlateau,
                {"factor": 0.1},
            ]
        elif optimizer == "Scheduler-args":
            scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau
            scheduler_kwargs = {"factor": 1e-3}
            check_scheduler = [
                torch.optim.lr_scheduler.ReduceLROnPlateau,
                {"factor": 1e-3},
            ]
        met.setup(
            optimizer=optimizer,
            scheduler=scheduler,
            optimizer_kwargs=optimizer_kwargs,
            scheduler_kwargs=scheduler_kwargs,
        )
        met.synthesize(max_iter=5)
        if not isinstance(met.optimizer, check_optimizer[0]):
            raise ValueError("Didn't properly set optimizer!")
        state_dict = met.optimizer.state_dict()["param_groups"][0]
        for k, v in check_optimizer[1].items():
            if state_dict[k] != v:
                raise ValueError(
                    "Didn't properly set optimizer kwargs! "
                    f"Expected {v} but got {state_dict[k]}!"
                )
        if check_scheduler is not None:
            if not isinstance(met.scheduler, check_scheduler[0]):
                raise ValueError("Didn't properly set scheduler!")
            state_dict = met.scheduler.state_dict()
            for k, v in check_scheduler[1].items():
                if met.scheduler.state_dict()[k] != v:
                    raise ValueError("Didn't properly set scheduler kwargs!")
        elif met.scheduler is not None:
            raise ValueError("Didn't set scheduler to None!")

    @pytest.mark.skipif(DEVICE.type == "cpu", reason="Only makes sense to test on cuda")
    @pytest.mark.parametrize(
        "model", ["naive.Identity", "PortillaSimoncelli"], indirect=True
    )
    @pytest.mark.filterwarnings(
        "ignore:Validating whether model can work with coarse-to-fine:UserWarning"
    )
    def test_map_location(self, curie_img, model, tmp_path):
        if hasattr(model, "scales"):
            met = po.synth.MetamerCTF(curie_img, model)
        else:
            met = po.synth.Metamer(curie_img, model)
        met.synthesize(max_iter=4, store_progress=True)
        met.save(op.join(tmp_path, "test_metamer_map_location.pt"))
        # calling load with map_location effectively switches everything
        # over to that device
        model.to("cpu")
        if hasattr(model, "scales"):
            met_copy = po.synth.MetamerCTF(curie_img.to("cpu"), model)
        else:
            met_copy = po.synth.Metamer(curie_img.to("cpu"), model)
        # end up with slightly different outputs on gpu vs cpu
        met_copy.load(
            op.join(tmp_path, "test_metamer_map_location.pt"),
            map_location="cpu",
            tensor_equality_atol=1e-6,
        )
        assert met_copy.metamer.device.type == "cpu"
        assert met_copy.image.device.type == "cpu"
        met_copy.synthesize(max_iter=4, store_progress=True)
        # reset model device for other tests
        model.to(DEVICE)

    @pytest.mark.skipif(DEVICE.type == "cpu", reason="Only makes sense to test on cuda")
    @pytest.mark.parametrize(
        "model", ["naive.Identity", "PortillaSimoncelli"], indirect=True
    )
    @pytest.mark.filterwarnings(
        "ignore:Validating whether model can work with coarse-to-fine:UserWarning"
    )
    def test_to_midsynth(self, curie_img, model):
        if hasattr(model, "scales"):
            met = po.synth.MetamerCTF(curie_img, model)
        else:
            met = po.synth.Metamer(curie_img, model)
        met.synthesize(max_iter=4, store_progress=2)
        assert met.metamer.device.type == "cuda"
        assert met.image.device.type == "cuda"
        met.to("cpu")
        met.synthesize(max_iter=4, store_progress=2)
        assert met.metamer.device.type == "cpu"
        assert met.image.device.type == "cpu"
        met.to("cuda")
        met.synthesize(max_iter=4, store_progress=2)
        assert met.metamer.device.type == "cuda"
        assert met.image.device.type == "cuda"

    @pytest.mark.parametrize(
        "model",
        [
            "naive.Identity",
            "NonModule",
            "frontend.OnOff.nograd",
            "frontend.OnOff.nograd.ctf",
        ],
        indirect=True,
    )
    @pytest.mark.parametrize("to_type", ["dtype", "device"])
    @pytest.mark.filterwarnings("ignore:Unable to call model.to:UserWarning")
    @pytest.mark.filterwarnings(
        "ignore:Validating whether model can work with coarse-to-fine:UserWarning"
    )
    def test_to(self, curie_img, model, to_type):
        if hasattr(model, "scales"):
            met = po.synth.MetamerCTF(curie_img, model)
        else:
            met = po.synth.Metamer(curie_img, model)
        met.synthesize(max_iter=5)
        if to_type == "dtype":
            met.to(torch.float64)
            assert met.image.dtype == torch.float64
            assert met.metamer.dtype == torch.float64
        # can only run this one if we're on a device with CPU and GPU.
        elif to_type == "device" and DEVICE.type != "cpu":
            met.to("cpu")
        met.metamer - met.image
        met.synthesize(max_iter=5)
        # reset so we don't mess up further tests
        if to_type == "dtype":
            met.to(torch.float32)
        elif to_type == "device" and DEVICE.type != "cpu":
            met.to(DEVICE)

    # this determines whether we mix across channels or treat them separately,
    # both of which are supported
    @pytest.mark.parametrize("model", ["ColorModel", "naive.Identity"], indirect=True)
    def test_multichannel(self, model, color_img):
        met = po.synth.Metamer(color_img, model)
        met.synthesize(max_iter=5)
        assert met.metamer.shape == color_img.shape, (
            "Metamer image should have the same shape as input!"
        )

    # this determines whether we mix across batches (e.g., a video model) or
    # treat them separately, both of which are supported
    @pytest.mark.parametrize("model", ["VideoModel", "naive.Identity"], indirect=True)
    def test_multibatch(self, model, einstein_img, curie_img):
        img = torch.cat([curie_img, einstein_img], dim=0)
        met = po.synth.Metamer(img, model)
        met.synthesize(max_iter=5)
        assert met.metamer.shape == img.shape, (
            "Metamer image should have the same shape as input!"
        )

    # we assume that the target representation has no gradient attached, so
    # doublecheck that (validate_model should ensure this)
    @pytest.mark.parametrize("model", ["frontend.OnOff.nograd"], indirect=True)
    def test_rep_no_grad(self, model, einstein_img):
        met = po.synth.Metamer(einstein_img, model)
        assert met.target_representation.grad is None, (
            "Target representation has a gradient attached, how?"
        )
        met.synthesize(max_iter=5)
        assert met.target_representation.grad is None, (
            "Target representation has a gradient attached, how?"
        )

    @pytest.mark.parametrize("model", ["frontend.OnOff.nograd"], indirect=True)
    def test_nan_loss(self, model, einstein_img):
        # clone to prevent NaN from showing up in other tests
        img = einstein_img.clone()
        met = po.synth.Metamer(img, model)
        met.synthesize(max_iter=5)
        met.target_representation[..., 0, 0] = torch.nan
        with pytest.raises(ValueError, match="Found a NaN in loss during optimization"):
            met.synthesize(max_iter=1)

    @pytest.mark.parametrize("model", ["naive.Identity"], indirect=True)
    def test_change_precision_save_load(self, model, einstein_img, tmp_path):
        # Identity model doesn't change when you call .to() with a dtype
        # (unlike those models that have weights) so we use it here
        met = po.synth.Metamer(einstein_img, model)
        met.synthesize(max_iter=5)
        met.to(torch.float64)
        assert met.metamer.dtype == torch.float64, "dtype incorrect!"
        met.save(op.join(tmp_path, "test_metamer_change_prec_save_load.pt"))
        met_copy = po.synth.Metamer(einstein_img.to(torch.float64), model)
        met_copy.load(op.join(tmp_path, "test_metamer_change_prec_save_load.pt"))
        met_copy.synthesize(max_iter=5)
        assert met_copy.metamer.dtype == torch.float64, "dtype incorrect!"

    @pytest.mark.filterwarnings("ignore:Loss has converged:UserWarning")
    @pytest.mark.parametrize(
        "model", ["frontend.OnOff.nograd", "PortillaSimoncelli"], indirect=True
    )
    @pytest.mark.filterwarnings(
        "ignore:Validating whether model can work with coarse-to-fine:UserWarning"
    )
    def test_stop_criterion(self, einstein_img, model):
        # checking that this hits the criterion and stops early, so set seed
        # for reproducibility
        po.tools.set_seed(0)
        if hasattr(model, "scales"):
            met = po.synth.MetamerCTF(einstein_img, model)
            synth_kwargs = {
                "change_scale_criterion": None,
                "ctf_iters_to_check": 2,
                "max_iter": 200,
            }
            stop_crit = 1e-1
        else:
            met = po.synth.Metamer(einstein_img, model)
            synth_kwargs = {"max_iter": 35}
            stop_crit = 1e-5
        # takes different numbers of iter to converge on GPU and CPU
        met.synthesize(stop_criterion=stop_crit, stop_iters_to_check=5, **synth_kwargs)
        # losses[-1] corresponds to the *current* loss (of met.metamer), not the loss
        # from the most recent iteration. so losses[-2] is the loss of the last
        # synthesis iteration.
        assert abs(met.losses[-6] - met.losses[-2]) < stop_crit, (
            "Didn't stop when hit criterion!"
        )
        assert abs(met.losses[-7] - met.losses[-3]) > stop_crit, (
            "Stopped after hit criterion!"
        )



"""
Fetch data using pooch.

This is inspired by scipy's datasets module.
"""

import pathlib
import sys

from tqdm.auto import tqdm

__all__ = ["DOWNLOADABLE_FILES", "fetch_data"]


def __dir__() -> list[str]:
    return __all__


REGISTRY = {
    "plenoptic-test-files.tar.gz": "a6b8e03ecc8d7e40c505c88e6c767af5da670478d3bebb4e13a9d08ee4f39ae8",  # noqa: E501
    "ssim_images.tar.gz": "19c1955921a3c37d30c88724fd5a13bdbc9620c9e7dfaeaa3ff835283d2bb42e",  # noqa: E501
    "ssim_analysis.mat": "921d324783f06d1a2e6f1ce154c7ba9204f91c569772936991311ff299597f24",  # noqa: E501
    "msssim_images.tar.gz": "a01273c95c231ba9e860dfc48f2ac8044ac3db13ad7061739c29ea5f9f20382c",  # noqa: E501
    "MAD_results.tar.gz": "29794ed7dc14626f115b9e4173bff88884cb356378a1d4f1f6cd940dd5b31dbe",  # noqa: E501
    "portilla_simoncelli_matlab_test_vectors.tar.gz": "83087d4d9808a3935b8eb4197624bbae19007189cd0d786527084c98b0b0ab81",  # noqa: E501
    "portilla_simoncelli_test_vectors.tar.gz": "d67787620a0cf13addbe4588ec05b276105ff1fad46e72f8c58d99f184386dfb",  # noqa: E501
    "portilla_simoncelli_images.tar.gz": "4d3228fbb51de45b4fc81eba590d20f5861a54d9e46766c8431ab08326e80827",  # noqa: E501
    "portilla_simoncelli_synthesize.npz": "9c304580cd60e0275a2ef663187eccb71f2e6a31883c88acf4c6a699f4854c80",  # noqa: E501
    "portilla_simoncelli_synthesize_torch_v1.12.0.npz": "5a76ef223bac641c9d48a0b7f49b3ce0a05c12a48e96cd309866b1e7d5e4473f",  # noqa: E501
    "portilla_simoncelli_synthesize_gpu.npz": "324efc2a6c54382aae414d361c099394227b56cd24460eebab2532f70728c3ee",  # noqa: E501
    "portilla_simoncelli_scales.npz": "eae2db6bd5db7d37c28d8f8320c4dd4fa5ab38294f5be22f8cf69e5cd5e4936a",  # noqa: E501
    "sample_images.tar.gz": "0ba6fe668a61e9f3cb52032da740fbcf32399ffcc142ddb14380a8e404409bf5",  # noqa: E501
    "test_images.tar.gz": "eaf35f5f6136e2d51e513f00202a11188a85cae8c6f44141fb9666de25ae9554",  # noqa: E501
    "tid2013.tar.gz": "bc486ac749b6cfca8dc5f5340b04b9bb01ab24149a5f3a712f13e9d0489dcde0",  # noqa: E501
    "portilla_simoncelli_test_vectors_refactor.tar.gz": "b72661836e5830c1473b8a2292075a8e9c1aca00faf97cc6809ec28f19d3f9ce",  # noqa: E501
    "portilla_simoncelli_synthesize_torch_v1.12.0_ps-refactor.npz": "9525844b71cf81509b86ed9677172745353588c6bb54e4de8000d695598afa47",  # noqa: E501
    "portilla_simoncelli_synthesize_gpu_ps-refactor.npz": "9fbb490f1548133f6aa49c54832130cf70f8dc6546af59688ead17f62ab94e61",  # noqa: E501
    "portilla_simoncelli_scales_ps-refactor.npz": "ce11d85e6bcf5fad1b819c36dac584c3e933706a0ee423ea1c76ffe0daccbae5",  # noqa: E501
    "portilla_simoncelli_synthesize_torch_v1.12.0_ps-refactor-2.npz": "ffd967543d58a03df390008c35878791590520624aa0e5e5a26ad3f877345ab4",  # noqa: E501
    "example_eigendistortion.pt": "87080836713e8efe1e7ff29538099e82a26b8700080e1bc1d30f00de1a54b2f5",  # noqa: E501
    "load_image_test.tar.gz": "8a2b92dc0d442695c45b1e908ef0a04cae35c5f21b774a93b9fc6b675423b526",  # noqa: E501
    "berardino_onoff.pt": "2174a40005489b9c94acc91213b2f6d57a75f262caf118cb1980658eadbfd047",  # noqa: E501
    "berardino_vgg16.pt": "5e0d10f4a367244879cd8a61c453992370ab801db1f66e10caa1ee2ecfab8ca4",  # noqa: E501
    "ps_regression.tar.gz": "8419f2df2ab9856e5a2f698afff73e4cab76faf65818c3c334f0141078a40914",  # noqa: E501
}

OSF_TEMPLATE = "https://osf.io/download/{}"
# these are all from the OSF project at https://osf.io/ts37w/.
REGISTRY_URLS = {
    "plenoptic-test-files.tar.gz": OSF_TEMPLATE.format("q9kn8"),
    "ssim_images.tar.gz": OSF_TEMPLATE.format("j65tw"),
    "ssim_analysis.mat": OSF_TEMPLATE.format("ndtc7"),
    "msssim_images.tar.gz": OSF_TEMPLATE.format("5fuba"),
    "MAD_results.tar.gz": OSF_TEMPLATE.format("jwcsr"),
    "portilla_simoncelli_matlab_test_vectors.tar.gz": OSF_TEMPLATE.format("qtn5y"),
    "portilla_simoncelli_test_vectors.tar.gz": OSF_TEMPLATE.format("8r2gq"),
    "portilla_simoncelli_images.tar.gz": OSF_TEMPLATE.format("eqr3t"),
    "portilla_simoncelli_synthesize.npz": OSF_TEMPLATE.format("a7p9r"),
    "portilla_simoncelli_synthesize_torch_v1.12.0.npz": OSF_TEMPLATE.format("gbv8e"),
    "portilla_simoncelli_synthesize_gpu.npz": OSF_TEMPLATE.format("tn4y8"),
    "portilla_simoncelli_scales.npz": OSF_TEMPLATE.format("xhwv3"),
    "sample_images.tar.gz": OSF_TEMPLATE.format("6drmy"),
    "test_images.tar.gz": OSF_TEMPLATE.format("au3b8"),
    "tid2013.tar.gz": OSF_TEMPLATE.format("uscgv"),
    "portilla_simoncelli_test_vectors_refactor.tar.gz": OSF_TEMPLATE.format("ca7qt"),
    "portilla_simoncelli_synthesize_torch_v1.12.0_ps-refactor.npz": OSF_TEMPLATE.format(
        "vmwzd"
    ),
    "portilla_simoncelli_synthesize_gpu_ps-refactor.npz": OSF_TEMPLATE.format("mqs6y"),
    "portilla_simoncelli_scales_ps-refactor.npz": OSF_TEMPLATE.format("nvpr4"),
    "portilla_simoncelli_synthesize_torch_v1.12.0_ps-refactor-2.npz": OSF_TEMPLATE.format(  # noqa: E501
        "en8du"
    ),
    "example_eigendistortion.pt": OSF_TEMPLATE.format("gwhz2"),
    "load_image_test.tar.gz": OSF_TEMPLATE.format("avpzq"),
    "berardino_onoff.pt": OSF_TEMPLATE.format("uqfa8"),
    "berardino_vgg16.pt": OSF_TEMPLATE.format("6r87b"),
    "ps_regression.tar.gz": OSF_TEMPLATE.format("7t4fj/?revision=8"),
}

#: List of files that can be downloaded using :func:`fetch_data`
DOWNLOADABLE_FILES = list(REGISTRY_URLS.keys())

try:
    import pooch
except ImportError:
    pooch = None
    retriever = None
else:
    retriever = pooch.create(
        # Use the default cache folder for the operating system
        # Pooch uses appdirs (https://github.com/ActiveState/appdirs) to
        # select an appropriate directory for the cache on each platform.
        path=pooch.os_cache("plenoptic"),
        base_url="",
        urls=REGISTRY_URLS,
        registry=REGISTRY,
        retry_if_failed=2,
        allow_updates="POOCH_ALLOW_UPDATES",
        env="PLENOPTIC_CACHE_DIR",
    )


def _find_shared_directory(paths: list[pathlib.Path]) -> pathlib.Path:
    """
    Find directory shared by all paths.

    Helper function for when downloading tar archives.

    Parameters
    ----------
    paths
        List of paths to check.

    Returns
    -------
    shared_dir
        Most recent common ancestor.
    """
    for dir in paths[0].parents:
        if all([dir in p.parents for p in paths]):
            break
    return dir


def fetch_data(dataset_name: str) -> pathlib.Path:
    """
    Download data, using pooch. These are largely used for testing.

    To view list of downloadable files, look at :const:`DOWNLOADABLE_FILES`.

    This checks whether the data already exists and is unchanged and downloads
    again, if necessary. If dataset_name ends in .tar.gz, this also
    decompresses and extracts the archive, returning the Path to the resulting
    directory. Else, it just returns the Path to the downloaded file.

    Parameters
    ----------
    dataset_name
        Name of the dataset to download.

    Returns
    -------
    path
        Path of the downloaded dataset.

    Raises
    ------
    ImportError
        If ``pooch`` isn't installed.

    Examples
    --------
    .. plot::

      >>> import plenoptic as po
      >>> from plenoptic.data import fetch
      >>> path = fetch.fetch_data("portilla_simoncelli_images.tar.gz")
      >>> len(list(path.glob("*")))
      38
      >>> img = po.load_images(path / "fig3b.jpg")
      >>> po.imshow(img)
      <PyrFigure size ...>
    """
    if retriever is None:
        raise ImportError(
            "Missing optional dependency 'pooch'."
            " Please use pip or "
            "conda to install 'pooch'."
        )
    processor = pooch.Untar() if dataset_name.endswith(".tar.gz") else None
    use_ascii = bool(sys.platform == "win32")
    fname = retriever.fetch(
        dataset_name,
        progressbar=tqdm(
            total=1,
            ncols=79,
            unit_scale=True,
            delay=1e-5,
            leave=True,
            unit="B",
            ascii=use_ascii,
        ),
        processor=processor,
    )
    if dataset_name.endswith(".tar.gz"):
        fname = _find_shared_directory([pathlib.Path(f) for f in fname])
    else:
        fname = pathlib.Path(fname)
    return fname
